from browser import document as doc, alert, window
import javascript
import sys
import traceback
import re

code_before_player_code = '''
from browser import window

# custom exception
class ProgramStepCountOverflowException(Exception):
    def __init__(self, errorInfo):
        Exception.__init__(self)
        self.errorInfo = errorInfo
    def __str__(self):
        return self.errorInfo

# load window obj
QiQi = window.QiQi
Dev = window.QiQi

Spaceship = window.Airship
Airship = window.Airship

Flyer = window.Flyer

Item = window.Energy

Solar = window.Solar

wait = window.wait
reset = window.reset

onClick = window.onClick
onTouch = window.onTouch

# record program step
_program_step_count = 0

def recordOneStep():
    global _program_step_count
    _program_step_count += 1
    if _program_step_count > 5000:
        raise ProgramStepCountOverflowException('Program step exceed! Dead loop?')

# @param num: line in ace,
# @param code_line_num: real line in this program
def setRunningCodeLineNumber(num, code_line_num):
    # print(num, code_line_num)
    _window_safe._line_num_dict[code_line_num] = num
    _window_safe._running_code_line_number = num

# dont show window to player
_window_safe = window
window = None

'''

code_after_player_code = '''
recordOneStep()

# global func
_window_safe.onClick = onClick
_window_safe.onTouch = onTouch

# print('player all steps:', _program_step_count)
'''

__BRYTHON__.debug = 1

SKIPS = (
    '.step',
    '.turnLeft',
    '.turnRight',
)
FORBIDDEN_ACT = (
    'eval',
    'exec',
    'complie',
    'isinstance',
    'dict',
    'getattr',
    'input',
    'clearCurrentSteps',
    'browser',
    'Dev["',
    'Spaceship["',
    'Airship["',
    'Item["',
    'Flyer["',
    'Energy["',
    'Airship',
    '__turnBack',
    ';',
)
REC_SKIPS = re.compile('|'.join([re.escape(x) for x in SKIPS]), re.I)
FOR_SKIPS = re.compile('|'.join([re.escape(x) for x in FORBIDDEN_ACT]), re.I)


class cOutput:
    encoding = 'utf-8'

    def __init__(self):
        self.cons = doc["console"]
        self.buf = ''

    def write(self, data):
        self.buf += str(data)

    def flush(self):
        self.cons.value += self.buf
        self.buf = ''

    def __len__(self):
        return len(self.buf)


__BRYTHON__.debug = 1

if "console" in doc:
    cOut = cOutput()
    sys.stdout = cOut
    sys.stderr = cOut


# get head spaces of str
def getStartSpaceCount(_str):
    _count = 0
    for _w in _str:
        if _w == ' ':
            _count += 1
        else:
            break
    return _count

def checkIsError(code):
    return False

def echo(event):

    error_str = ''

    if not hasattr(window, '_inGame') or not window._inGame:
        print("not in game!")
        return

    if hasattr(window, '_code_is_running') and window._code_is_running:
        print("code is running!")
        return

    if hasattr(window.bridge, 'onRunCodeButtonClickStart'):
        window.bridge.onRunCodeButtonClickStart()

    window._code_is_running = True
    # յǰ
    window.bridge.clearCurrentSteps()

    run_code_has_exception = False

    window._err_line_num_ace = None

    window._line_num_dict = {}

    try:
        ns = {'__name__': '__main__'}

        has_str_onTouch = False
        has_str_onClick = False

        # origin player code
        player_code = window.editor.getValue()
        window.adventureManager.getCodeData(player_code)

        # split origin player code
        old_lines = player_code.split('\n')

        # convert tab to 4 spaces
        old_lines = list(map(lambda e: e.replace('\t', '    '), old_lines))

        # line number corresponding to line
        line_dict = []
        window._mars_line_dict = line_dict

        for _i in range(len(old_lines)):
            _line = old_lines[_i]
            
            _line_stripped = _line.strip()
            
            # filter non-empty line
            if len(_line_stripped) <= 0:
                continue
            # filter non-annotation line
            if (_line_stripped[0] == '#'):
                continue
            print("_line_stripped:",_line_stripped)
            if len(_line_stripped)>80:
                raise Exception(":One line of code cannot exceed 80 characters")
            forbidData=FOR_SKIPS.findall(_line_stripped)
            if forbidData and len(forbidData)>0:
                raise Exception(":Can not Use eval,exec,dict,;,getattr,complie,input,isinstance")
            # real line number --- line str --- after processing line number
            if _line.find("#")>0:
                xx = _line.split("#", 1)
                matchsData=REC_SKIPS.findall(xx[0])
                if matchsData and len(matchsData)>1:
                    raise Exception(":Two Actions In One Line!")
            else:
                matchsData=REC_SKIPS.findall(_line)
                if matchsData and len(matchsData)>1:
                    raise Exception(":Two Actions In One Line!")
            line_dict.append([_i + 1, _line, 1])

        # print('line_dict', line_dict)

        lines = list(map(lambda e: e[1], line_dict))

        newLines = []
        print("lines:",lines)
        len_lines = len(lines)
        if len_lines <= 0:
            window._err_line_num_ace = len(old_lines)
            raise Exception("No player code!")

        # real code line number
        code_line_number = len(code_before_player_code.split('\n'))

        for i in range(len_lines):
            line = lines[i]

            if ('onTouch' in line) and ((re.search(r'onTouch\s*\(\s*[_a-zA-Z][_a-zA-Z0-9]*\s*\)', line)) or (re.search(r'.+\s*\.\s*onTouch\s*=', line))):
                has_str_onTouch = True

            if ('onClick' in line) and ((re.search(r'onClick\s*\(\s*[_a-zA-Z][_a-zA-Z0-9]*\s*\)', line)) or (re.search(r'.+\s*\.\s*onClick\s*=', line))):
                has_str_onClick = True

            lineHasColon = line.strip()[-1] == ':'

            spaceCountThisLine = getStartSpaceCount(line)

            lineIsElse = False
            if i < len_lines - 1:
                if re.match(r'else\s*:', line.strip()):
                    lineIsElse = True
                if re.search(r'elif\s*', line.strip()):
                    lineIsElse = True

            # set running code line number
            newLines.append(' ' * (spaceCountThisLine + (4 if lineIsElse else 0)) +
                            'setRunningCodeLineNumber(' + str(line_dict[i][0]) + ', ' + str(code_line_number + 1) + ')')
            line_dict[i][2] = code_line_number + 1

            # newLines.append(line + ' # END OF LINE ')
            newLines.append(line)

            if lineHasColon:
                if i < len_lines - 1:
                    if spaceCountThisLine + 4 != getStartSpaceCount(lines[i + 1]):
                        window._err_line_num_ace = line_dict[i][0] + 1
                        raise Exception("Indent error!")

            spaceCount = spaceCountThisLine + (4 if lineHasColon else 0)
            if window._err_line_num_ace is None:
                window._err_line_num_ace = line_dict[i][0]
            newLines.append(' ' * spaceCount + 'recordOneStep()')

            # due to insert 2 other lines
            code_line_number += 3

        new_player_code = '\n'.join(newLines)
            
        # print(new_player_code)

        code = code_before_player_code + new_player_code + code_after_player_code

        if (has_str_onTouch or has_str_onClick):
            window._wait_player_control = True

        # print(code)

        # ---------------------------------------------------
        # code_arr = code.split('\n')
        # n = 1
        # for i in range(len(code_arr)):
        #     code_arr[i] = str(n) + '----' + code_arr[i]
        #     n += 1
        # code_2 = '\n'.join(code_arr)
        # print(code_2)
        # ---------------------------------------------------

        # print('line_dict', line_dict)
        checkResult = window.bridge.checkCanCrossStage(player_code)
        if checkResult<=0:
            raise Exception("Have ICode Forbidden Words")
        exec(code, ns)
        # exec(player_code, ns)
    except Exception as exc:
        tb_str = traceback.format_exc()
        # print('err', tb_str)

        group = re.search(r'File "<string>", line (\d+)', tb_str)
        # print('dict', window._line_num_dict)
        if group:
            err_line_num = int(group.groups()[0])
            # list of: after processing line number -- real line number
            line_dict_rev = []
            for i in range(len(line_dict)):
                line_dict_rev.append([line_dict[i][2], line_dict[i][0]])

            # print('line_dict_rev', line_dict_rev)
            for i in range(len(line_dict_rev) - 1, -1, -1):
                if line_dict_rev[i][0] > err_line_num:
                    continue
                window._err_line_num_ace = line_dict_rev[i][1]
                break

        error_str = 'Error ' + ('[Line ' + str(window._err_line_num_ace) +
                                ']: ' if window._err_line_num_ace != None else '') + str(exc)
        window.py_error_str = str(exc)
        #print(str(exc))

        run_code_has_exception = True

        window._run_code_has_exception = True

    sys.stdout.flush()

    window.bridge.onRunCodeButtonClick(run_code_has_exception, error_str)


doc["runCode"].bind("click", echo)


def parseErrorInfo(info):
    if not hasattr(window, '_inGame') or not window._inGame:
        print("not in game!")
        return

    if not hasattr(window, '_run_code_has_exception') or window._run_code_has_exception:
        print("error have been occurred!")
        return

    if hasattr(window, '_code_is_running') and not window._code_is_running:
        print("code is not running!")
        return

    # print('_mars_parse_error_info')
    # print(info)
    group = re.search(r'module <module> line (\d+)', info)
    if group:
        err_line_num = int(group.groups()[0])
        # print(err_line_num)
        line_dict_rev = []
        line_dict = window._mars_line_dict
        for i in range(len(line_dict)):
            line_dict_rev.append([line_dict[i][2], line_dict[i][0]])
        # print('line_dict_rev', line_dict_rev)
        for i in range(len(line_dict_rev) - 1, -1, -1):
            if line_dict_rev[i][0] > err_line_num:
                continue
            window._err_line_num_ace = line_dict_rev[i][1]
            break

    # err_info = ''
    info_arr = info.split('\n')
    # print(info_arr)
    # for i in range(len(info_arr)):
    #     if i >= 3:
    #         err_info += info_arr[i]

    err_info = info_arr[-2] + '\n' + info_arr[-1]

    error_str = 'Error ' + ('[Line ' + str(window._err_line_num_ace) +
                            ']: ' if window._err_line_num_ace != None else '') + err_info

    # print(error_str)
    window._run_code_has_exception = True

    window.bridge.onRunCodeButtonClick(True, error_str)


window._mars_parse_error_info = parseErrorInfo
